#include <map>
#include <list>
#include <string>

#include "autoptr"

typedef autoptr<class XMLTree> XMLTreeP;
class XMLTree: public ptrable
{
public:
    // tags
    typedef std::multimap<std::string, XMLTreeP> tagmap_t;
    tagmap_t tags;
    // current tag params
    typedef std::map<std::string, std::string> parammap_t;
    parammap_t params;
    // tag value
    std::string value;
    
    class XMLTreeSet operator[] (const std::string&) const;

    template<typename Func>
    void ForEach(const std::string& key, Func& f) const;
    
    void Dump(unsigned indent=0) const;
};

const XMLTree ParseXML(const std::string& s, bool KeepNS = true);

class XMLTreeSet
{
public:
    const XMLTreeSet operator[] (const std::string&) const;
    
    operator const std::string () const;
    //operator const XMLTreeP () const;

    template<typename Func>
    void ForEach(const std::string& key, Func& f) const
    {
        for(std::list<XMLTreeP>::const_iterator
            i = matching_trees.begin();
            i != matching_trees.end();
            ++i)
        {
            const XMLTree& p = *(*i);
            p.ForEach(key, f);
        }
    }
    
    template<typename Func>
    void IterateStrings(Func& func) const;
    
    template<typename Func>
    void IterateTrees(Func& func) const;
    
    void Dump() const;
    
    bool empty() const { return matching_strings.empty() && matching_trees.empty(); }

protected:
    friend class XMLTree;
    void AddTree(XMLTreeP p);
    void SetValue(const std::string& s);
    void Combine(const XMLTreeSet& b);

private:
    std::list<XMLTreeP> matching_trees;
    std::list<std::string> matching_strings;
};

template<typename Func>
void XMLTree::ForEach(const std::string& key, Func& f) const
{
    (*this)[key].IterateStrings(f);
}

template<typename Func>
void XMLTreeSet::IterateStrings(Func& func) const
{
    std::for_each(matching_strings.begin(), matching_strings.end(), func);
}

template<typename Func>
void XMLTreeSet::IterateTrees(Func& func) const
{
    std::list<XMLTreeP>::const_iterator i;
    for(i=matching_trees.begin(); i!=matching_trees.end(); ++i)
        func(**i);
}
